import pytest

from rmediator.decorators import request, request_handler


class TestRequestHandler:
    def test__request_handler__no_request_attribute__raises_value_error(self):
        class InvalidRequestHandler:
            pass

        with pytest.raises(ValueError):
            request_handler(InvalidRequestHandler)

    def test__request_handler__no_response_attribute__raises_value_error(
        self, get_valid_request
    ):
        request_class = get_valid_request

        class InvalidRequestHandler:
            request = request_class

        with pytest.raises(ValueError):
            request_handler(InvalidRequestHandler)

    def test__request_handler__no_handle_method__raises_value_error(
        self, get_valid_request
    ):
        request_class, response_class = get_valid_request, get_valid_request.response

        class InvalidRequestHandler:
            request = request_class
            response = response_class

        with pytest.raises(ValueError):
            request_handler(InvalidRequestHandler)

    def test__request_handler__request_type_mismatch__raises_value_error(
        self, get_valid_request
    ):
        request_class, response_class = get_valid_request, get_valid_request.response

        class InvalidRequestHandler:
            request = request_class
            response = response_class

            def handle(self, request: get_valid_request.response):
                pass

        with pytest.raises(ValueError):
            request_handler(InvalidRequestHandler)

    def test__request_handler__response_type_mismatch__raises_value_error(
        self, get_valid_request
    ):
        request_class, response_class = get_valid_request, get_valid_request.response

        class InvalidRequestHandler:
            request = request_class
            response = response_class

            def handle(self, request: get_valid_request) -> get_valid_request:
                pass

        with pytest.raises(ValueError):
            request_handler(InvalidRequestHandler)


@pytest.fixture
def get_valid_request(get_response):
    @request
    class Request:
        response = get_response

    return Request


@pytest.fixture
def get_response():
    class Response:
        pass

    return Response



def request(cls):
    if "response" not in cls.__dict__:
        raise ValueError(
            f"Request class cannot be registered, {cls} must contain attribute 'response' to signify the response expected for this request."
        )

    cls.rmediator_decorated_request = True

    return cls



import pytest

from rmediator.decorators import request


class TestRequest:
    def test__request__invalid_no_response__raises_value_error(self):
        with pytest.raises(ValueError):

            @request
            class InvalidRequest:
                pass

    def test__request__valid(self):

        @request
        class ValidRequest:
            response = str



from abc import ABCMeta, abstractmethod
from typing import TypeVar

TRequest = TypeVar("TRequest")
TResponse = TypeVar("TResponse")


class RequestHandler(metaclass=ABCMeta):
    request = TRequest
    response = TResponse

    @abstractmethod
    def handle(self, request: TRequest) -> TResponse:
        pass



from threading import Lock
from typing import Any, Dict

from rmediator.types import Request, RequestHandler


class SingletonMeta(type):
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class Mediator(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self.__request_handlers: Dict[type, RequestHandler] = {}

    def send(self, request: Request) -> Any:
        handler = self.__request_handlers.get(type(request))
        if not handler:
            raise ValueError(
                f"Request cannot be handled, No handler has been registered for {type(request)}."
            )

        return handler.handle(request)

    def register_handler(
        self, request_type: type[Request], handler: RequestHandler
    ) -> None:
        self.__check_request_validity(request_type)
        self.__check_request_handler_validity(type(handler))

        if request_type in self.__request_handlers:
            raise ValueError(
                f"Handler cannot be registered, Another handler for request type {request_type} has already been registered. Check {type(self.__request_handlers[request_type])}."
            )

        if handler.response != request_type.response:  # type: ignore
            raise ValueError(
                f"Handler cannot be registered, Handler response {handler.response} does not match request response {request_type.response}."
            )

        self.__request_handlers[request_type] = handler

    def __check_request_validity(self, request_type: type[Request]) -> None:
        if not hasattr(request_type, "rmediator_decorated_request"):
            raise ValueError(
                f"Handler cannot be registered, The request {request_type} class has not been decorated using @request."
            )

    def __check_request_handler_validity(
        self, handler_type: type[RequestHandler]
    ) -> None:
        if not hasattr(handler_type, "rmediator_decorated_request_handler"):
            raise ValueError(
                f"Handler cannot be registered, The request handler {type(handler_type)} class has not been decorated using @request_handler."
            )



def request_handler(cls):
    if "request" not in cls.__dict__:
        raise ValueError(
            f"Request handler class cannot be registered, {cls} must contain attribute 'request' to signify the reequest it handles."
        )

    if "response" not in cls.__dict__:
        raise ValueError(
            f"Request handler class cannot be registered, {cls} must contain attribute 'response' to signify the response it sends after handling request."
        )

    if "handle" not in cls.__dict__:
        raise ValueError(
            f"Request handler class cannot be registered, {cls} must contain method 'handle' that handles the request."
        )

    if cls.request is None:  # type: ignore
        raise ValueError(
            f"Request handler class cannot be registered, {cls} must contain attribute 'request' to signify the request it handles."
        )

    if cls.handle.__annotations__.get("request", None) != cls.request:  # type: ignore
        raise ValueError(
            f"Request handler class cannot be registered, {cls} handle method must have 'request' parameter of type {cls.request}."  # type: ignore
        )

    if cls.handle.__annotations__.get("return", None) != cls.response:  # type: ignore
        raise ValueError(
            f"Request handler class cannot be registered, {cls} handle method must have return type of {cls.response}."  # type: ignore
        )

    cls.rmediator_decorated_request_handler = True

    return cls



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rmediator",
    author="Fikernew Birhanu",
    author_email="fikernew.birhanu.waju@gmail.com",
    description="Request Mediator PyPI (Python Package Index) Package",
    keywords="rmediator, pypi, package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ffekirnew/rmediator",
    project_urls={
        "Documentation": "https://github.com/ffekirnew/rmediator",
        "Bug Reports": "https://github.com/ffekirnew/rmediator/issues",
        "Source Code": "https://github.com/ffekirnew/rmediator",
        # 'Funding': '',
        # 'Say Thanks!': '',
    },
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        # see https://pypi.org/classifiers/
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    # install_requires=['Pillow'],
    extras_require={
        "dev": ["check-manifest"],
        # 'test': ['coverage'],
    },
    # entry_points={
    #     'console_scripts': [  # This can provide executable scripts
    #         'run=examplepy:main',
    # You can execute `run` in bash to run `main()` in src/examplepy/__init__.py
    #     ],
    # },
)



import pytest

from rmediator import Mediator
from rmediator.decorators import request, request_handler


class TestMediator:
    def test__mediator__valid_registration(
        self, mediator, get_valid_request, get_valid_request_handler
    ):
        mediator.register_handler(get_valid_request, get_valid_request_handler())

    def test__mediator__invalid_types_given_to_register_funciton__raises_value_error(
        self, mediator
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(str, bool)

    def test__mediator__invalid_duplicate_registration__raises_value_error(
        self, mediator, get_valid_request, get_valid_request_handler
    ):
        mediator.register_handler(get_valid_request, get_valid_request_handler())

        with pytest.raises(ValueError):
            mediator.register_handler(get_valid_request, get_valid_request_handler())


@pytest.fixture
def mediator():
    return Mediator()


@pytest.fixture
def get_valid_request(get_response):
    @request
    class Request:
        response = get_response

    return Request


@pytest.fixture
def get_response():
    class Response:
        pass

    return Response


@pytest.fixture
def get_valid_request_handler(get_valid_request, get_response):
    @request_handler
    class RequestHandler:
        request = get_valid_request
        response = get_response

        def handle(self, request: get_valid_request) -> get_response:
            pass

    return RequestHandler
